/*
 * ModularEEG firmware for one-way transmission, v0.5.4-p2
 * Copyright (c) 2002-2003, Joerg Hansmann, Jim Peters, Andreas Robinson
 * License: GNU General Public License (GPL) v2
 * Compiles with AVR-GCC v3.3.
 *
 * Note: -p2 in the version number means this firmware is for packet version 2.
 */

//////////////////////////////////////////////////////////////

/*

////////// Packet Format Version 2 ////////////

// 17-byte packets are transmitted from the ModularEEG at 256Hz,
// using 1 start bit, 8 data bits, 1 stop bit, no parity, 57600 bits per second.

// Minimial transmission speed is 256Hz * sizeof(modeeg_packet) * 10 = 43520 bps.

struct modeeg_packet
{
	uint8_t		sync0;		// = 0xa5
	uint8_t		sync1;		// = 0x5a
	uint8_t		version;	// = 2
	uint8_t		count;		// packet counter. Increases by 1 each packet.
	uint16_t	data[6];	// 10-bit sample (= 0 - 1023) in big endian (Motorola) format.
	uint8_t		switches;	// State of PD5 to PD2, in bits 3 to 0.
};

// Note that data is transmitted in big-endian format.
// By this measure together with the unique pattern in sync0 and sync1 it is guaranteed, 
// that re-sync (i.e after disconnecting the data line) is always safe.

// At the moment communication direction is only from Atmel-processor to PC.
// The hardware however supports full duplex communication. This feature
// will be used in later firmware releases to support the PWM-output and
// LED-Goggles.

*/

//////////////////////////////////////////////////////////////

/*
 * Program flow:
 *
 * When 256Hz timer expires: goto SIGNAL(SIG_OVERFLOW0)
 * SIGNAL(SIG_OVERFLOW0) enables the ADC
 *
 * Repeat for each channel in the ADC: 
 * Sampling starts. When it completes: goto SIGNAL(SIG_ADC)
 * SIGNAL(SIG_ADC) reads the sample and restarts the ADC.
 *
 * SIGNAL(SIG_ADC) writes first byte to UART data register
 * (UDR) which starts the transmission over the serial port.
 *
 * Repeat for each byte in packet:
 * When transmission begins and UDR empties: goto SIGNAL(SIG_UART_DATA)
 *
 * Start over from beginning.
 */

// =====================================================
// Last Built with WinAVR-20100110
// Supported processors: ATmega8, ATmega16 and AT90S4434 
// =====================================================

#include <avr/io.h>
#include <inttypes.h>
#include <avr/interrupt.h>
#include <compat/deprecated.h>
//#include <avr/signal.h>

#define Debug_Led_On			PORTB |= 0x20; 		DDRB |= 0x20;
#define Debug_Led_Off			PORTB &= (~0x20); 	DDRB |= 0x20;

#if defined (__AVR_ATmega328P__)
	#define Crystal_Freq_16MHz		16	//16MHz external crystal oscillator - especially for Olimexino 328!
#else
	#undef  Crystal_Freq_16MHz		// In all other cases 7.3728MHz crystal is used!
#endif

#define NUMCHANNELS 6
#define HEADERLEN 4
#define PACKETLEN (NUMCHANNELS * 2 + HEADERLEN + 1)

#define SAMPFREQ 256
#define TIMER0VAL 256 - ((7372800 / 256) / SAMPFREQ)

//char const channel_order[]= { 0, 3, 1, 4, 2, 5 };
char const channel_order[]= { 0, 1, 2, 3, 4, 5 };

/** The transmission packet */

volatile uint8_t TXBuf[PACKETLEN];

/** Next byte to read or write in the transmission packet. */

volatile uint8_t TXIndex;  

/** Current channel being sampled. */

volatile uint8_t CurrentCh;



/** Sampling timer (timer 0) interrupt handler */

//SIGNAL(SIG_OVERFLOW0)
ISR(TIMER0_OVF_vect)
{

	//Debug_Led_On;
    outb(TCNT0, TIMER0VAL); //Reset timer to get correct sampling frequency.

    CurrentCh = 0;

    // Write header and footer:

    // Increase packet counter (fourth byte in header)

    TXBuf[3]++;

    //Get state of switches on PD2..5, if any (last byte in packet).

    TXBuf[2 * NUMCHANNELS + HEADERLEN] = (inp(PIND) >> 2) &0x0F;


#if defined (__AVR_ATmega8__)
	
	cbi(UCSRB, UDRIE);      //Ensure UART IRQ's are disabled.
    sbi(ADCSR, ADIF);       //Reset any pending ADC interrupts
    sbi(ADCSR, ADIE);       //Enable ADC interrupts.

#elif defined (__AVR_ATmega16__)
	
	cbi(UCSRB, UDRIE);      //Ensure UART IRQ's are disabled.
	sbi(ADCSRA, ADIF);      //Reset any pending ADC interrupts
    sbi(ADCSRA, ADIE);      //Enable ADC interrupts.

#elif defined (__AVR_ATmega328P__)
	
	cbi(UCSR0B, UDRIE0);     //Ensure UART IRQ's are disabled.
	sbi(ADCSRA, ADIF);       //Reset any pending ADC interrupts
    sbi(ADCSRA, ADIE);       //Enable ADC interrupts.
	sbi(ADCSRA, ADSC) ;		 // Start conversion!!!!!!!!!!!!!!!!!!!

#else//(__AVR_AT90S4434__)

	cbi(UCR, UDRIE);      //Ensure UART IRQ's are disabled.
    sbi(ADCSR, ADIF);       //Reset any pending ADC interrupts
    sbi(ADCSR, ADIE);       //Enable ADC interrupts.

#endif

    //The ADC will start sampling automatically as soon
    //as sleep is executed in the main-loop.
	//Debug_Led_On;
}



/** AD-conversion-complete interrupt handler. */

//SIGNAL(SIG_ADC)
ISR(ADC_vect)
{
    volatile uint8_t i;

    i = 2 * CurrentCh + HEADERLEN;

    TXBuf[i+1] = inp(ADCL);
    TXBuf[i] = inp(ADCH);

    CurrentCh++;
	//Debug_Led_On;

    if (CurrentCh < NUMCHANNELS)
    {
        outb(ADMUX, (channel_order[CurrentCh]));  //Select the next channel.
        //The next sampling is started automatically.
#if defined (__AVR_ATmega328P__)
		sbi(ADCSRA, ADSC) ;		 // Start conversion!!!!!!!!!!!!!!!!!!!
#endif
    }
    else
    {
        outb(ADMUX, channel_order[0]);      //Prepare next conversion, on channel 0.

        // Disable ADC interrupts to prevent further calls to SIG_ADC.
#if defined (__AVR_ATmega8__) | defined (__AVR_AT90S4434__)

        cbi(ADCSR, ADIE);

#elif defined (__AVR_ATmega16__) | defined (__AVR_ATmega328P__)

		cbi(ADCSRA, ADIE);

#endif

        // Hand over to SIG_UART_DATA, by starting
        // the UART transfer and enabling UDR IRQ's.
#if 	defined (__AVR_ATmega328P__)
		outb(UDR0, TXBuf[0]);		
#else	
		outb(UDR, TXBuf[0]);
#endif
        
		 
#if defined (__AVR_ATmega8__) | defined (__AVR_ATmega16__)

    	sbi(UCSRB, UDRIE);

#elif defined (__AVR_ATmega328P__)
		
		sbi(UCSR0B, UDRIE0);

#else//(__AVR_AT90S4434__)

		sbi(UCR, UDRIE);

#endif
        TXIndex = 1;
		//Debug_Led_On;
    }
}


/*** UART data transmission register-empty interrupt handler ***/

#if defined	(__AVR_ATmega328P__)
	ISR(USART_UDRE_vect)		/* USART, Data Register Empty interrupt */
#else
	SIGNAL(SIG_UART_DATA)
#endif
{

#if defined	(__AVR_ATmega328P__)
	//Debug_Led_On;
	outb(UDR0, TXBuf[TXIndex]);  //Send next byte
#else
    outb(UDR, TXBuf[TXIndex]);  //Send next byte
#endif

    TXIndex++;
    if (TXIndex == PACKETLEN)   //See if we're done with this packet
    {
#if defined (__AVR_ATmega8__) | defined (__AVR_ATmega16__)

    	cbi(UCSRB, UDRIE);      //Disable SIG_UART_DATA interrupts.
                                //Next interrupt will be a SIG_OVERFLOW0.

#elif defined (__AVR_ATmega328P__)

		cbi(UCSR0B, UDRIE0);    //Disable SIG_UART_DATA interrupts.
                                //Next interrupt will be a SIG_OVERFLOW0.

#else//(__AVR_AT90S4434__)

		cbi(UCR, UDRIE);		//Disable SIG_UART_DATA interrupts.
                                //Next interrupt will be a SIG_OVERFLOW0.
#endif
        
    }
}

/** Initialize PWM output (PB1 = 14Hz square wave signal) */

void pwm_init(void)
{
	// Set timer/counter 1 to use 10-bit PWM mode.
	// The counter counts from zero to 1023 and then back down
	// again. Each time the counter value equals the value
	// of OCR1(A), the output pin is toggled.
	// The counter speed is set in TCCR1B, to clk / 256 = 28800Hz.
	// Effective frequency is then clk / 256 / 2046 = 14 Hz

#if defined (__AVR_ATmega8__) | defined (__AVR_ATmega16__) | defined (__AVR_ATmega328P__)

    outb(OCR1AH,2);	// Set OCR1A = 512
    outb(OCR1AL,0);
    outb(TCCR1A, ((1<<COM1A1) + (1<<WGM11) + (1<<WGM10))); // Set 10-bit PWM mode
    outb(TCCR1B, (1 << CS12));	// Start and let run at clk / 256 Hz.

#else // __AVR_AT90S4434__

    outb(OCR1AH,2);	// Set OCR1 = 512
    outb(OCR1AL,0);
    outb(TCCR1A, ((1<<COM1A1) + (1<<PWM11) + (1<<PWM10))); // Set 10-bit PWM mode
    outb(TCCR1B, (1 << CS12));	// Start and let run at clk / 256 Hz.

#endif
}

int main( void )
{

#if defined	(Crystal_Freq_16MHz)
	//Devide system clock by two to acheive the same performance and time intervals
	sbi(CLKPR,CLKPCE);	// Clock Prescaler Change Enable
	outb(CLKPR,0x01);	// Set prescaller to 1:2
#endif	

    //Write packet header and footer

    TXBuf[0] = 0xa5;        //Sync 0
    TXBuf[1] = 0x5a;        //Sync 1
    TXBuf[2] = 2;           //Protocol version
    TXBuf[3] = 0;           //Packet counter

    //Set up the ports.
#if defined (__AVR_ATmega8__) | defined (__AVR_ATmega328P__)

    outb(DDRD,  0xc2);
    outb(DDRB,  0x07);
    outb(PORTD, 0xff);
    outb(PORTB, 0xff);

#elif defined (__AVR_ATmega16__)

	DDRD |= 0x22;			// Enable PWM output (PD5/OC1A) and TxD output
	PORTD |= 0x22;			// Set outputs to high

#endif


    //Select sleep mode = idle.
#if defined (__AVR_ATmega8__) | defined (__AVR_ATmega16__) | defined (__AVR_ATmega328P__)

	outb(MCUCR,(inp(MCUCR) | (1<<SE)) & (~(1<<SM0) | ~(1<<SM1) | ~(1<<SM2)));


#else // __AVR_AT90S4434__

	outb(MCUCR,(inp(MCUCR) | (1<<SE)) & (~(1<<SM0)) & (~(1<<SM1)) );

#endif

    //Initialize the ADC

    // Timings for sampling of one 10-bit AD-value:
    //
    // prescaler > ((XTAL / 200kHz) = 36.8 => 
    // prescaler = 64 (ADPS2 = 1, ADPS1 = 1, ADPS0 = 0)
    // ADCYCLE = XTAL / prescaler = 115200Hz or 8.68 us/cycle
    // 14 (single conversion) cycles = 121.5 us (8230 samples/sec)
    // 26 (1st conversion) cycles = 225.69 us

    outb(ADMUX, 0);         //Select channel 0

    //Prescaler = 64, free running mode = off, interrupts off.

#if defined (__AVR_ATmega8__) | defined (__AVR_AT90S4434__)

    outb(ADCSR, ((1<<ADPS2) | (1<<ADPS1)));
    sbi(ADCSR, ADIF);		//Reset any pending ADC interrupts
    sbi(ADCSR, ADEN);		//Enable the ADC

#elif defined (__AVR_ATmega16__) | defined (__AVR_ATmega328P__)

	outb(ADCSRA, ((1<<ADPS2) | (1<<ADPS1)));
    sbi(ADCSRA, ADIF);		//Reset any pending ADC interrupts
    sbi(ADCSRA, ADEN);		//Enable the ADC
	
#endif

    //Initialize the UART

#if defined (__AVR_ATmega8__) | defined (__AVR_ATmega16__)

	outb(UBRRH,0);              //Set speed to 57600 bps
    outb(UBRRL,7);
	outb(UCSRA, 0);
	outb(UCSRC, ((1<<URSEL) | (1<<UCSZ1) | (1<<UCSZ0)));	// 8 bits data length
    outb(UCSRB, (1<<TXEN));									// Transmitter Enabled

#elif defined (__AVR_ATmega328P__)

	#if defined	(Crystal_Freq_16MHz)
		outb(UBRR0, 8);               //Set speed to 57600 bps
    	outb(UCSR0B, (1<<TXEN0));
	#else
		outb(UBRR0, 7);               //Set speed to 57600 bps
    	outb(UCSR0B, (1<<TXEN0));
	#endif
    

#else // __AVR_AT90S4434__

    outb(UBRR, 7);               //Set speed to 57600 bps
    outb(UCR, (1<<TXEN));

#endif

    //Initialize timer 0  -> used to sample ADC  

#if defined (__AVR_ATmega328P__)
	outb(TCNT0, 0);             //Clear it.
    outb(TCCR0B, 4);            //Start it. Frequency = clk / 256
    outb(TIMSK0, (1<<TOIE0));   //Enable the interrupts.
#else
	outb(TCNT0, 0);             //Clear it.
    outb(TCCR0, 4);             //Start it. Frequency = clk / 256
    outb(TIMSK, (1<<TOIE0));    //Enable the interrupts.
#endif

    //Initialize PWM (optional)
	
    pwm_init();

	//while(1){	Debug_Led_On; outb(UDR0, 0x55);	}	// Just for debug purpose!

    sei();	// Enable all interrupts

    //Now, we wait. This is an event-driven program, so nothing much
    //happens here.

    while (1)
    {
        __asm__ __volatile__ ("sleep");
    }
}
